#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdlib.h>
#include "Strings.h"

static const char SepTbl[SEPARATE_TABLE_SIZE] = {
	0,0,0,0,0,0,0,0, 0,2,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,
	2,0,0,0,0,0,0,0, 1,1,0,0,1,0,0,0, // (),
	0,0,0,0,0,0,0,0, 0,0,1,1,0,1,0,0, //:;=
	0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0, 0,0,0,1,0,1,0,0, //[]
	0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0, 0,0,0,1,0,1,0,0, //{}
};


static int get_utf8_lead_bytes(char ch)
{
	if(ch == '\0') return 0;
 
	if((ch & 0x80) == 0) return 1;

	int pos = 0;
	for( char tmp = ch & 0xfc; (tmp & 0x80); tmp = tmp << 1 )
	{
		++pos;
	}
	return pos;
}

//---------------------------------------------------------------------------
//  StringSeparator class
//---------------------------------------------------------------------------
StringSeparator::StringSeparator()
{
	mCharCode = CharCode_Ansi;
	scan = NULL;
	MaxScan = CurCount = 0;
	mSeparateBuffer = nullptr;
	mSeparateBufferSize = 0;

	StringChar = '"';
	memcpy(SeparateTable, SepTbl, SEPARATE_TABLE_SIZE);
}

StringSeparator::~StringSeparator()
{
	if(mSeparateBuffer != nullptr)
		free(mSeparateBuffer);
	if(scan != NULL){
		for(int i=0; i<MaxScan; i++){
			if(scan[i] != NULL) free(scan[i]);
		}
		free(scan);
	}
}

StringSeparator::CharacterCode StringSeparator::GetCharacterCode() const
{
	return mCharCode;
}

void StringSeparator::SetCharacterCode(CharacterCode code)
{
	mCharCode = code;
}

void StringSeparator::GetSeparateTable(char *table) const
{
	memcpy(table, SeparateTable, SEPARATE_TABLE_SIZE);
}

void StringSeparator::SetSeparateTable(const char *table)
{
	memcpy(SeparateTable, table, SEPARATE_TABLE_SIZE);
}

char StringSeparator::GetSeparateTableValue(int index) const
{
	if(index >= 0 && index < SEPARATE_TABLE_SIZE)
		return SeparateTable[index];
	return 0;
}

void StringSeparator::SetSeparateTableValue(int index, char value)
{
	if(index >= 0 && index < SEPARATE_TABLE_SIZE)
		SeparateTable[index] = value;
}

char StringSeparator::GetStringChar(void) const
{
	return StringChar;
}

void StringSeparator::SetStringChar(char ch)
{
	StringChar = ch;
}

bool StringSeparator::SetScan(int index, const char *str)
{
	if(index >= MaxScan){
		scan = (char **)realloc(scan, sizeof(char*) * (index+1));
		for(int i=MaxScan; i<=index; i++)
			scan[i] = NULL;
		MaxScan = index+1;
	}

	if(scan[index] != NULL)
		free(scan[index]);

	size_t len = strlen(str) + 1;
	scan[index] = (char *)malloc(len);
	memcpy(scan[index], str, len);

	return true;
}

bool StringSeparator::SetScan(int index, const wchar_t *wstr)
{
	int len = WideCharToMultiByte(CP_OEMCP, 0, wstr, -1, NULL, 0, NULL, NULL);
	char *mchar = (char *)malloc(len);
	WideCharToMultiByte(CP_OEMCP, 0, wstr, -1, mchar, len, NULL, NULL);

	SetScan(index, mchar);

	free(mchar);

	return true;
}

int StringSeparator::Separate(const char *str)
{
	bool instr = false;
	int utf8len;

	size_t maxlen = strlen(str) + 1;
	if(mSeparateBuffer == nullptr){
		mSeparateBuffer = (char*)malloc(maxlen);
		if(mSeparateBuffer == nullptr){
			CurCount = 0;
			return 0;
		}
		mSeparateBufferSize = maxlen;
	}else if(maxlen > mSeparateBufferSize){
		char *buf = (char*)realloc(mSeparateBuffer, maxlen);
		if(buf == nullptr){
			CurCount = 0;
			return 0;
		}
		mSeparateBuffer = buf;
		mSeparateBufferSize = maxlen;
	}

	char *buf = mSeparateBuffer;
	size_t len = 0;
	int count = 0;
	for(const char *ptr = str; len < maxlen; ptr++)
	{
		if(mCharCode == CharCode_Ansi && ::IsDBCSLeadByte(*ptr)){
			buf[len++] = *ptr++;
			buf[len++] = *ptr;
		}else if(mCharCode == CharCode_Utf8 && (utf8len = get_utf8_lead_bytes(*ptr)) > 1){
			for(int n=1; n<utf8len; n++){
				buf[len++] = *ptr++;
			}
			buf[len++] = *ptr;
		} else switch(*ptr){
		case 0x0D:
		case 0x0A:
		case 0:
			if(len > 0){
				buf[len] = '\0';
				SetScan(count++, buf);
			}
			goto GS_EXIT;
		default:
			if(*ptr == StringChar){
				if(instr){
					buf[len] = '\0';
					SetScan(count++, buf);
					len=0; instr=false;
				} else {
					if(len > 0){
						buf[len] = '\0';
						SetScan(count++, buf);
						len=0;
					}
					instr=true;
				}
			} else if(*ptr>=0 && SeparateTable[*ptr]){
				switch(SeparateTable[*ptr]){
				case Separator_Symbol:
					if(instr) goto ADDCHAR;
					if(len > 0){
						buf[len] = '\0';
						SetScan(count++, buf);
						len=0;
					}
					if(SeparateTable[*ptr] == 1){
						buf[0] = *ptr;
						buf[1] = '\0';
						SetScan(count++, buf);
					}
					break;
				case Separator_Delimiter: // space, tab and so on
					if(instr) goto ADDCHAR;
					if(len > 0){
						buf[len] = '\0';
						SetScan(count++, buf);
						len=0;
					}
					break;
				}
			} else {
ADDCHAR:
				buf[len++] = *ptr;
			}
		}
	}

GS_EXIT:
	CurCount = count;
	return CurCount;
}

bool StringSeparator::CheckCommand(const char *comstr, int optc)
//optc    count of option strings to need
{
	return (optc<=CurCount && _stricmp(scan[0], comstr) == 0);
}

bool StringSeparator::CheckString(int index, const char *str)
//optc    count of option strings to need
{
	return (index<CurCount && _stricmp(scan[index], str) == 0);
}

int StringSeparator::ToInt(int index) const
{
	if(index >= CurCount) return 0;
	return atoi(scan[index]);
}

int StringSeparator::ToHex(int index) const
{
	if(index >= CurCount) return 0;
	int value=0;
	sscanf_s(scan[index], "%x", &value);
	return value;
}

double StringSeparator::ToDouble(int index) const
{
	if(index >= CurCount) return 0;
	return atof(scan[index]);
}


//---------------------------------------------------------------------------
//  FileStringSeparator class
//---------------------------------------------------------------------------
LineStringSeparator::LineStringSeparator() : StringSeparator()
{
	mLineBuffer = nullptr;
	mLineBufferSize = 0;
	mRawLine = nullptr;
}

LineStringSeparator::~LineStringSeparator()
{
	if(mLineBuffer != nullptr)
		free(mLineBuffer);
}

const char *LineStringSeparator::GetRawLine()
{
	return mRawLine;
}

//---------------------------------------------------------------------------
//  FileStringSeparator class
//---------------------------------------------------------------------------
FileStringSeparator::FileStringSeparator(FILE *_fh) : LineStringSeparator()
{
	fh = _fh;
}

FileStringSeparator::~FileStringSeparator()
{
}

int FileStringSeparator::GetLine(void)
//if file is end, return EOF, or count of elements
{
	__int64 fp;

	if(mLineBuffer == nullptr){
		const size_t s = 4096;
		mLineBuffer = (char*)malloc(s);
		if(mLineBuffer == nullptr){
			CurCount=EOF;
			return EOF;
		}
		mLineBufferSize = s;
	}
	for(size_t i=0; ; i++)
	{
		if(i >= mLineBufferSize-4){
			const size_t s = mLineBufferSize * 2;
			char *buf = (char*)realloc(mLineBuffer, s);
			if(buf == nullptr){
				CurCount=EOF;
				return EOF;
			}
			mLineBuffer = buf;
			mLineBufferSize = s;
		}
		char *buf = mLineBuffer;
		int c = fgetc(fh);
		switch(c){
		case EOF:
		GET_EOF:
			buf[i] = 0;
			if(i == 0){
				CurCount=EOF;
				goto EXIT;
			}
			goto LOAD_END;
		case 0x0D:
		GET_0D:
			buf[i] = 0;
			fp = _ftelli64(fh);
			c = fgetc(fh);
			if(c != 0x0A)
				_fseeki64(fh, fp, SEEK_SET);
			goto LOAD_END;
		case 0x0A:
		GET_0A:
			buf[i] = 0;
			goto LOAD_END;
		default:
			if(mCharCode == CharCode_Ansi){
				if(::IsDBCSLeadByte((BYTE)c)){
					buf[i++] = (BYTE)c;
					c = fgetc(fh);
					switch(c){
					case EOF: goto GET_EOF;
					case 0x0D: goto GET_0D;
					case 0x0A: goto GET_0A;
					}
				}
				buf[i] = (BYTE)c;
			}else if(mCharCode == CharCode_Utf8){
				int len = get_utf8_lead_bytes(c);
				for(int n=1; n<len; n++){
					buf[i++] = (BYTE)c;
					c = fgetc(fh);
					if(c == EOF) goto GET_EOF;
				}
				buf[i] = (BYTE)c;
			}
			break;
		}
	}

LOAD_END:
	Separate(mLineBuffer);
	mRawLine = mLineBuffer;
EXIT:
	return CurCount;
}

//---------------------------------------------------------------------------
//  AsciiFileString class
//---------------------------------------------------------------------------
AsciiFileString::AsciiFileString(FILE *_fh) : FileStringSeparator(_fh)
{
	mLineCount = 0;
}

int AsciiFileString::GetLine(void)
//if file is end, return EOF, or count of elements
{
	if(mLineBuffer == nullptr){
		const size_t s = 8192;
		mLineBuffer = (char*)malloc(s);
		if(mLineBuffer == nullptr){
			CurCount=EOF;
			return EOF;
		}
		mLineBufferSize = s;
	}
	char *buf = mLineBuffer;
	int buflen = (int)mLineBufferSize;
	bool eof = false;
	if(fgets(buf, buflen, fh) == NULL){
		eof = true;
	}else{
		while(1){
			int len = (int)strlen(buf);
			if(len != buflen-1) break;
			if(buf[buflen-2] == 0x0A || buf[buflen-2] == 0x0D) break;

			buf = (char*)realloc(mLineBuffer, mLineBufferSize + 8192);
			if(buf == nullptr)
				return 0;
			mLineBuffer = buf;
			mLineBufferSize += 8192;
			if(fgets(buf + buflen-1, 8192+1, fh) == NULL){
				eof = true;
				break;
			}
			buflen += 8192;
		}
	}
	if(!eof){
		if(mLineCount == 0 && ((BYTE)buf[0] == 0xEF && (BYTE)buf[1] == 0xBB && (BYTE)buf[2] == 0xBF)){
			mCharCode = CharCode_Utf8;
			CurCount = Separate(buf + 3);
			mRawLine = buf + 3;
		}else{
			CurCount = Separate(buf);
			mRawLine = buf;
		}
	}else{
		CurCount = EOF;
	}

	mLineCount++;
	return CurCount;
}

//---------------------------------------------------------------------------
//  MacFileString class
//---------------------------------------------------------------------------
MacFileString::MacFileString(FILE *_fh) : FileStringSeparator(_fh)
{
}

int MacFileString::GetLine(void)
//if file is end, return EOF, or count of elements
{
	return FileStringSeparator::GetLine();
}


